# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""LAMBADA dataset."""

import os

from etils import epath
import tensorflow_datasets.public_api as tfds

_LAMBADA_DATASET_URL = (
    'https://zenodo.org/record/2630551/files/lambada-dataset.tar.gz?download=1'
)


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for lambada dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(
            {
                'passage': tfds.features.Text(),
            }
        ),
        supervised_keys=None,
        homepage='https://zenodo.org/record/2630551#.X4Xzn5NKjUI',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""

    dl_dir = dl_manager.download_and_extract(_LAMBADA_DATASET_URL)

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                'filepath': os.path.join(
                    dl_dir, 'lambada_development_plain_text.txt'
                )
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                'filepath': os.path.join(dl_dir, 'lambada_test_plain_text.txt')
            },
        ),
    ]

  def _generate_examples(self, filepath):
    """Yields examples."""
    with epath.Path(filepath).open() as f:
      for idx, line in enumerate(f):
        key = '%s_%d' % (os.path.basename(filepath), idx)
        yield key, {'passage': line}



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BeeDataset dataset."""
from __future__ import annotations

import json
import random

import numpy as np
import tensorflow_datasets.public_api as tfds


class BeeDatasetConfig(tfds.core.BuilderConfig):
  """BuilderConfig for the BeeDataset.

  Args:
  image_width (int): Desired image width.
  image_height (int): Desired image heigth.
  """

  def __init__(self, image_height=300, image_width=150, **kwargs):
    super().__init__(**kwargs)
    self.width = image_width
    self.height = image_height
    self.depth = 3


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for BeeDataset dataset."""

  VERSION = tfds.core.Version('1.0.0')

  URL = 'https://raspbee.de/BeeDataset_20201121.zip'

  BEE_CFG_300 = BeeDatasetConfig(
      name='bee_dataset_300',
      description='BeeDataset images with 300 pixel height and 150 pixel width',
      version='1.0.0',
      image_height=300,
      image_width=150,
  )

  BEE_CFG_200 = BeeDatasetConfig(
      name='bee_dataset_200',
      description='BeeDataset images with 200 pixel height and 100 pixel width',
      version='1.0.0',
      image_height=200,
      image_width=100,
  )

  BEE_CFG_150 = BeeDatasetConfig(
      name='bee_dataset_150',
      description='BeeDataset images with 200 pixel height and 100 pixel width',
      version='1.0.0',
      image_height=150,
      image_width=75,
  )

  BUILDER_CONFIGS = [BEE_CFG_300, BEE_CFG_200, BEE_CFG_150]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    t_shape = (
        self.builder_config.height,
        self.builder_config.width,
        self.builder_config.depth,
    )
    features = tfds.features.FeaturesDict({
        'input': tfds.features.Image(shape=t_shape),
        'output': {
            'varroa_output': np.float64,
            'pollen_output': np.float64,
            'wasps_output': np.float64,
            'cooling_output': np.float64,
        },
    })

    return self.dataset_info_from_configs(
        features=features,
        supervised_keys=('input', 'output'),
        homepage='https://raspbee.de',
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(self.URL)
    return {
        'train': self._generate_examples(path),
    }

  def _generate_examples(self, path):
    # Load labels and image path.
    data = json.loads((path / 'data.json').read_text())
    indexes = list(data.keys())
    random.shuffle(indexes)
    for name in indexes:
      labels = []
      entry = data[name]

      for lbl in ['varroa', 'pollen', 'wasps', 'cooling']:
        labels.append(1.0 if entry[lbl] else 0.0)

      img = path / f'images_{self.builder_config.height}' / name

      yield name + str(self.builder_config.height), {
          'input': img,
          'output': {
              'varroa_output': labels[0],
              'pollen_output': labels[1],
              'wasps_output': labels[2],
              'cooling_output': labels[3],
          },
      }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for unnatural_instructions dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.text.unnatural_instructions import unnatural_instructions


class UnnaturalInstructionsTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for unnatural_instructions dataset."""

  DATASET_CLASS = unnatural_instructions.UnnaturalInstructions
  SPLITS = {
      'train': 14,  # Number of fake train example
  }


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for croissant_builder."""

from __future__ import annotations

import dataclasses
import re
import typing

from tensorflow_datasets.core.utils import conversion_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import mlcroissant as mlc

if typing.TYPE_CHECKING:
  # pylint: disable=g-bad-import-order
  import mlcroissant as mlc

_HUGGINGFACE_URL_PREFIX = "https://huggingface.co/datasets/"
_VERSION_REGEX_WITHOUT_PATCH = re.compile(r"^(?P<major>\d+)\.(?P<minor>\d+)$")


@dataclasses.dataclass(frozen=True)
class SplitReference:
  """Information about a split reference in a Croissant dataset."""

  # A split record set in a Croissant dataset.
  split_record_set: mlc.RecordSet
  # A field from another record set that references split_record_set.
  reference_field: mlc.Field


def get_croissant_version(version: str | None) -> str | None:
  """Returns the possibly corrected Croissant version in TFDS format.

  TFDS expects versions to follow the Semantic versioning 2.0.0 syntax, but
  Croissant is more lax and accepts also {major.minor}. To avoid raising errors
  in these cases, we add a `0` as a patch version to the Croissant-provided
  version.

  Args:
    version: The Croissant version.

  Returns:
    The Croissant version in TFDS format.
  """
  if not version:
    return None
  if _VERSION_REGEX_WITHOUT_PATCH.match(version):
    return f"{version}.0"
  return version


def extract_localized_string(
    attribute: str | dict[str, str] | None,
    language: str | None = None,
    field_name: str = "text field",
) -> str | None:
  """Returns the text in the specified language from a potentially localized object.

  Some attributes in Croissant (e.g., `name` and `description`) can be
  localized, meaning that they can be either simple strings, or dictionaries
  mapping language codes to strings (e.g., `{"en": "English Name", "fr": "Nom
  français"}`). This function extracts the text in the specified language from a
  potentially localized object.

  Args:
    attribute: The object containing the text, which can be a simple string, a
      dictionary mapping language codes to strings, or None.
    language: The desired language code. If None, a heuristic is used: 'en' is
      preferred, otherwise the first available language in the dictionary.
    field_name: The name of the field being processed (e.g., "name",
      "description"), used for error messages.

  Returns:
    The text string in the desired language, or None if the input is None.

  Raises:
    ValueError: If the text_object is an empty dictionary, or if the specified
      language is not found.
    TypeError: If attribute is not a str, dict, or None.
  """
  if attribute is None:
    return None
  if isinstance(attribute, str):
    return attribute

  if not isinstance(attribute, dict):
    raise TypeError(
        f"{field_name} must be a string, dictionary, or None. Got"
        f" {type(attribute)}"
    )

  if language is None:
    # Try a heuristic language, e.g., 'en'.
    if "en" in attribute:
      return attribute["en"]
    # Otherwise, take the first language in the dict.
    try:
      first_lang = next(iter(attribute))
      return attribute[first_lang]
    except StopIteration as exc:
      raise ValueError(f"Dataset `{field_name}` dictionary is empty.") from exc
  elif language in attribute:
    return attribute[language]
  else:
    raise ValueError(
        f"Language '{language}' not found in {field_name} keys:"
        f" {list(attribute.keys())}."
    )


def get_dataset_name(dataset: mlc.Dataset, language: str | None = None) -> str:
  """Returns dataset name of the given MLcroissant dataset.

  Args:
    dataset: The MLcroissant dataset.
    language: For datasets with multiple names in different languages, this
      argument specifies the language to use.
  """
  if (url := dataset.metadata.url) and url.startswith(_HUGGINGFACE_URL_PREFIX):
    return url.removeprefix(_HUGGINGFACE_URL_PREFIX)
  name = extract_localized_string(
      dataset.metadata.name, language=language, field_name="name"
  )
  if name is None:
    # This case should ideally be prevented by mlcroissant's validation
    # ensuring metadata.name is not None.
    raise ValueError("Dataset name is missing.")
  return name


def get_tfds_dataset_name(
    dataset: mlc.Dataset, language: str | None = None
) -> str:
  """Returns TFDS compatible dataset name of the given MLcroissant dataset.

  Args:
    dataset: The MLcroissant dataset.
    language: For datasets with multiple names in different languages, this
      argument specifies the language to use.
  """
  dataset_name = get_dataset_name(dataset, language=language)
  return conversion_utils.to_tfds_name(dataset_name)


def get_record_set(
    tfds_config_name: str, metadata: mlc.Metadata
) -> mlc.RecordSet:
  """Returns the desired record set from a dataset's metadata."""
  for record_set in metadata.record_sets:
    if conversion_utils.to_tfds_name(record_set.id) == tfds_config_name:
      return record_set
  raise ValueError(
      f"Did not find any record set with the name {tfds_config_name}."
  )


def get_field(field_id: str, metadata: mlc.Metadata) -> mlc.Field:
  """Returns the desired field from a dataset's metadata."""
  for record_set in metadata.record_sets:
    for field in record_set.fields:
      if field.id == field_id:
        return field
  raise ValueError(f"Did not find any field with the name {field_id}.")


def get_record_set_for_field(
    field_id: str, metadata: mlc.Metadata
) -> mlc.RecordSet:
  """Given a field id, returns the record set it belongs to, if any."""
  for record_set in metadata.record_sets:
    for field in record_set.fields:
      if field.id == field_id:
        return record_set
  raise ValueError(f"Did not find any record set with field {field_id}.")


def get_split_recordset(
    record_set: mlc.RecordSet, metadata: mlc.Metadata
) -> SplitReference | None:
  """If a given recordset references a split recordset, returns it.

  Args:
    record_set: The record set to check.
    metadata: The metadata of the dataset.

  Returns:
    If found, a tuple containing: (the field referencing the split record set,
    and the split record set), None otherwise.
  """
  for field in record_set.fields:
    if field.references and field.references.field:
      # Check that the referenced record set is of type `cr:Split`.
      referenced_field = get_field(field.references.field, metadata)
      record_sets = [
          node
          for node in referenced_field.predecessors
          if isinstance(node, mlc.RecordSet)
      ]
      if not record_sets:
        raise ValueError(f"Field {field.id} has no RecordSet.")
      referenced_record_set = record_sets[0]
      if mlc.DataType.SPLIT in referenced_record_set.data_types:
        return SplitReference(referenced_record_set, field)
  return None


def get_record_set_ids(metadata: mlc.Metadata) -> list[str]:
  """Returns record set ids of the given MLcroissant metadata.

  Record sets which have the attribute `cr:Data` are excluded (e.g. splits that
  specify split or labels mappings).

  Args:
    metadata: The metadata of the dataset.
  """
  record_set_ids = []
  for record_set in metadata.record_sets:
    if record_set.data is not None:
      continue
    record_set_ids.append(record_set.id)
  return record_set_ids



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The rockyou dataset."""

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DOWNLOAD_URL = "https://github.com/brannondorsey/naive-hashcat/releases/download/data/rockyou.txt"


class Builder(tfds.core.GeneratorBasedBuilder):
  """This dataset contains passwords that were leaked or stolen from from various sites."""

  VERSION = tfds.core.Version("1.0.0")

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(
            {
                "password": tfds.features.Text(),
            }
        ),
        supervised_keys=None,
        homepage="https://wiki.skullsecurity.org/Passwords",
    )

  def _split_generators(self, dl_manager):
    dl_path = dl_manager.download(_DOWNLOAD_URL)
    return [
        tfds.core.SplitGenerator(
            name="train",
            gen_kwargs={
                "path": dl_path,
            },
        )
    ]

  def _generate_examples(self, path):
    with tf.io.gfile.GFile(path, "rb") as f:
      blines = f.readlines()

    for i, bline in enumerate(blines):
      yield i, {
          "password": bline.strip(),
      }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: smart_control_building.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1csmart_control_building.proto\x12#smart_buildings.smart_control.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\xfe\x01\n\x08ZoneInfo\x12\x0f\n\x07zone_id\x18\x01'
    b' \x01(\t\x12\x13\n\x0b\x62uilding_id\x18\x02'
    b' \x01(\t\x12\x18\n\x10zone_description\x18\x03'
    b' \x01(\t\x12\x0c\n\x04\x61rea\x18\x04'
    b' \x01(\x02\x12\x0f\n\x07\x64\x65vices\x18\x05'
    b' \x03(\t\x12I\n\tzone_type\x18\x06'
    b' \x01(\x0e\x32\x36.smart_buildings.smart_control.proto.ZoneInfo.ZoneType\x12\r\n\x05\x66loor\x18\x07'
    b' \x01(\x05"9\n\x08ZoneType\x12\r\n\tUNDEFINED\x10\x00\x12\x08\n\x04ROOM\x10\x01\x12\t\n\x05\x46LOOR\x10\x02\x12\t\n\x05OTHER\x10\n"\xa5\x07\n\nDeviceInfo\x12\x11\n\tdevice_id\x18\x01'
    b' \x01(\t\x12\x11\n\tnamespace\x18\x02'
    b' \x01(\t\x12\x0c\n\x04\x63ode\x18\x03'
    b' \x01(\t\x12\x0f\n\x07zone_id\x18\x04'
    b' \x01(\t\x12O\n\x0b\x64\x65vice_type\x18\x05'
    b' \x01(\x0e\x32:.smart_buildings.smart_control.proto.DeviceInfo.DeviceType\x12`\n\x11observable_fields\x18\x06'
    b' \x03(\x0b\x32\x45.smart_buildings.smart_control.proto.DeviceInfo.ObservableFieldsEntry\x12X\n\raction_fields\x18\x07'
    b' \x03(\x0b\x32\x41.smart_buildings.smart_control.proto.DeviceInfo.ActionFieldsEntry\x1ar\n\x15ObservableFieldsEntry\x12\x0b\n\x03key\x18\x01'
    b' \x01(\t\x12H\n\x05value\x18\x02'
    b' \x01(\x0e\x32\x39.smart_buildings.smart_control.proto.DeviceInfo.ValueType:\x02\x38\x01\x1an\n\x11\x41\x63tionFieldsEntry\x12\x0b\n\x03key\x18\x01'
    b' \x01(\t\x12H\n\x05value\x18\x02'
    b' \x01(\x0e\x32\x39.smart_buildings.smart_control.proto.DeviceInfo.ValueType:\x02\x38\x01"\xe7\x01\n\nDeviceType\x12\r\n\tUNDEFINED\x10\x00\x12\x07\n\x03\x46\x41N\x10\x01\x12\x07\n\x03PMP\x10\x02\x12\x07\n\x03\x46\x43U\x10\x03\x12\x07\n\x03VAV\x10\x04\x12\x06\n\x02\x44H\x10\x05\x12\x07\n\x03\x41HU\x10\x06\x12\x07\n\x03\x42LR\x10\x07\x12\x08\n\x04\x43\x44WS\x10\x08\x12\x06\n\x02\x43H\x10\t\x12\x08\n\x04\x43HWS\x10\n\x12\x06\n\x02\x43T\x10\x0b\x12\x06\n\x02\x44\x43\x10\x0c\x12\x07\n\x03\x44\x46R\x10\r\x12\x07\n\x03\x44MP\x10\x0e\x12\x07\n\x03HWS\x10\x0f\x12\x06\n\x02HX\x10\x10\x12\x07\n\x03MAU\x10\x11\x12\x07\n\x03SDC\x10\x12\x12\x06\n\x02UH\x10\x13\x12\x07\n\x03PWR\x10\x14\x12\x07\n\x03GAS\x10\x15\x12\x06\n\x02\x41\x43\x10\x16\x12\t\n\x05OTHER\x10\x17"w\n\tValueType\x12\x18\n\x14VALUE_TYPE_UNDEFINED\x10\x00\x12\x14\n\x10VALUE_CONTINUOUS\x10\x01\x12\x11\n\rVALUE_INTEGER\x10\x02\x12\x15\n\x11VALUE_CATEGORICAL\x10\x03\x12\x10\n\x0cVALUE_BINARY\x10\x04"G\n\x18SingleObservationRequest\x12\x11\n\tdevice_id\x18\x01'
    b' \x01(\t\x12\x18\n\x10measurement_name\x18\x02'
    b' \x01(\t"\xdf\x02\n\x19SingleObservationResponse\x12-\n\ttimestamp\x18\x01'
    b' \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x61\n\x1asingle_observation_request\x18\x02'
    b' \x01(\x0b\x32=.smart_buildings.smart_control.proto.SingleObservationRequest\x12\x19\n\x11observation_valid\x18\x03'
    b' \x01(\x08\x12\x1a\n\x10\x63ontinuous_value\x18\x04'
    b' \x01(\x02H\x00\x12\x17\n\rinteger_value\x18\x05'
    b' \x01(\x05H\x00\x12\x1b\n\x11\x63\x61tegorical_value\x18\x06'
    b' \x01(\tH\x00\x12\x16\n\x0c\x62inary_value\x18\x07'
    b' \x01(\x08H\x00\x12\x16\n\x0cstring_value\x18\x08'
    b' \x01(\tH\x00\x42\x13\n\x11observation_value"\xa7\x01\n\x12ObservationRequest\x12-\n\ttimestamp\x18\x01'
    b' \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x62\n\x1bsingle_observation_requests\x18\x02'
    b' \x03(\x0b\x32=.smart_buildings.smart_control.proto.SingleObservationRequest"\xf4\x01\n\x13ObservationResponse\x12-\n\ttimestamp\x18\x01'
    b' \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12H\n\x07request\x18\x02'
    b' \x01(\x0b\x32\x37.smart_buildings.smart_control.proto.ObservationRequest\x12\x64\n\x1csingle_observation_responses\x18\x03'
    b' \x03(\x0b\x32>.smart_buildings.smart_control.proto.SingleObservationResponse"\xd3\x01\n\x13SingleActionRequest\x12\x11\n\tdevice_id\x18\x01'
    b' \x01(\t\x12\x15\n\rsetpoint_name\x18\x02'
    b' \x01(\t\x12\x1a\n\x10\x63ontinuous_value\x18\x03'
    b' \x01(\x02H\x00\x12\x17\n\rinteger_value\x18\x04'
    b' \x01(\x05H\x00\x12\x1b\n\x11\x63\x61tegorical_value\x18\x05'
    b' \x01(\tH\x00\x12\x16\n\x0c\x62inary_value\x18\x06'
    b' \x01(\x08H\x00\x12\x16\n\x0cstring_value\x18\x07'
    b' \x01(\tH\x00\x42\x10\n\x0esetpoint_value"\xdd\x03\n\x14SingleActionResponse\x12I\n\x07request\x18\x01'
    b' \x01(\x0b\x32\x38.smart_buildings.smart_control.proto.SingleActionRequest\x12\x63\n\rresponse_type\x18\x02'
    b' \x01(\x0e\x32L.smart_buildings.smart_control.proto.SingleActionResponse.ActionResponseType\x12\x17\n\x0f\x61\x64\x64itional_info\x18\x03'
    b' \x01(\t"\xfb\x01\n\x12\x41\x63tionResponseType\x12\r\n\tUNDEFINED\x10\x00\x12\x0c\n\x08\x41\x43\x43\x45PTED\x10\x01\x12\x0b\n\x07PENDING\x10\x02\x12\r\n\tTIMED_OUT\x10\x03\x12\x1c\n\x18REJECTED_INVALID_SETTING\x10\x04\x12%\n!REJECTED_NOT_ENABLED_OR_AVAILABLE\x10\x05\x12\x15\n\x11REJECTED_OVERRIDE\x10\x06\x12\x1b\n\x17REJECTED_INVALID_DEVICE\x10\x07\x12\x1b\n\x17REJECTED_DEVICE_OFFLINE\x10\x08\x12\x0b\n\x07UNKNOWN\x10\t\x12\t\n\x05OTHER\x10\n"\x98\x01\n\rActionRequest\x12-\n\ttimestamp\x18\x01'
    b' \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12X\n\x16single_action_requests\x18\x02'
    b' \x03(\x0b\x32\x38.smart_buildings.smart_control.proto.SingleActionRequest"\xe0\x01\n\x0e\x41\x63tionResponse\x12-\n\ttimestamp\x18\x01'
    b' \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x43\n\x07request\x18\x02'
    b' \x01(\x0b\x32\x32.smart_buildings.smart_control.proto.ActionRequest\x12Z\n\x17single_action_responses\x18\x03'
    b' \x03(\x0b\x32\x39.smart_buildings.smart_control.proto.SingleActionResponseb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, 'smart_control_building_pb2', globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DEVICEINFO_OBSERVABLEFIELDSENTRY._options = None
  _DEVICEINFO_OBSERVABLEFIELDSENTRY._serialized_options = b'8\001'
  _DEVICEINFO_ACTIONFIELDSENTRY._options = None
  _DEVICEINFO_ACTIONFIELDSENTRY._serialized_options = b'8\001'
  _ZONEINFO._serialized_start = 103
  _ZONEINFO._serialized_end = 357
  _ZONEINFO_ZONETYPE._serialized_start = 300
  _ZONEINFO_ZONETYPE._serialized_end = 357
  _DEVICEINFO._serialized_start = 360
  _DEVICEINFO._serialized_end = 1293
  _DEVICEINFO_OBSERVABLEFIELDSENTRY._serialized_start = 712
  _DEVICEINFO_OBSERVABLEFIELDSENTRY._serialized_end = 826
  _DEVICEINFO_ACTIONFIELDSENTRY._serialized_start = 828
  _DEVICEINFO_ACTIONFIELDSENTRY._serialized_end = 938
  _DEVICEINFO_DEVICETYPE._serialized_start = 941
  _DEVICEINFO_DEVICETYPE._serialized_end = 1172
  _DEVICEINFO_VALUETYPE._serialized_start = 1174
  _DEVICEINFO_VALUETYPE._serialized_end = 1293
  _SINGLEOBSERVATIONREQUEST._serialized_start = 1295
  _SINGLEOBSERVATIONREQUEST._serialized_end = 1366
  _SINGLEOBSERVATIONRESPONSE._serialized_start = 1369
  _SINGLEOBSERVATIONRESPONSE._serialized_end = 1720
  _OBSERVATIONREQUEST._serialized_start = 1723
  _OBSERVATIONREQUEST._serialized_end = 1890
  _OBSERVATIONRESPONSE._serialized_start = 1893
  _OBSERVATIONRESPONSE._serialized_end = 2137
  _SINGLEACTIONREQUEST._serialized_start = 2140
  _SINGLEACTIONREQUEST._serialized_end = 2351
  _SINGLEACTIONRESPONSE._serialized_start = 2354
  _SINGLEACTIONRESPONSE._serialized_end = 2831
  _SINGLEACTIONRESPONSE_ACTIONRESPONSETYPE._serialized_start = 2580
  _SINGLEACTIONRESPONSE_ACTIONRESPONSETYPE._serialized_end = 2831
  _ACTIONREQUEST._serialized_start = 2834
  _ACTIONREQUEST._serialized_end = 2986
  _ACTIONRESPONSE._serialized_start = 2989
  _ACTIONRESPONSE._serialized_end = 3213
# @@protoc_insertion_point(module_scope)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python DataSource base class."""

from __future__ import annotations

from collections.abc import MappingView, Sequence
import dataclasses
from typing import Any, Callable


@dataclasses.dataclass(repr=False)
class PythonDataSource(MappingView, Sequence):
  """Python data source backed by Python objects: length and __getitem__."""

  length: int
  # If you have pickling issues for this function, define it in the upper scope
  # or get inspiration from _getitem in
  # tensorflow_datasets/core/features/dataset_feature.py.
  getitem: Callable[[int], Any]

  def __len__(self) -> int:
    return self.length

  def __iter__(self):
    for i in range(self.length):
      yield self[i]

  def __getitem__(self, i: int) -> Any:
    return self.getitem(i)

  def __repr__(self) -> str:
    return f'PythonDataSource(length={self.length})'



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datasets as hf_datasets
import huggingface_hub
import numpy as np
import pytest
from tensorflow_datasets.core import features as feature_lib
from tensorflow_datasets.core import registered
from tensorflow_datasets.core.utils import huggingface_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf


def test_convert_to_np_dtype_raises():
  with pytest.raises(TypeError, match='Unrecognized type.+'):
    huggingface_utils._convert_to_np_dtype('I am no dtype')


@pytest.mark.parametrize(
    'hf_dtype,np_dtype',
    [
        # from mapping
        ('bool', np.bool_),
        ('float', np.float32),
        ('double', np.float64),
        ('large_string', np.object_),
        ('utf8', np.object_),
        ('string', np.object_),
        ('binary', np.bytes_),
        # from np
        ('bool_', np.bool_),
        ('int32', np.int32),
        ('int64', np.int64),
        # from timestamp
        ('timestamp[s, tz=UTC]', np.int64),
        # from tf
        ('bfloat16', tf.bfloat16),
    ],
)
def test_convert_to_np_dtype(hf_dtype, np_dtype):
  assert huggingface_utils._convert_to_np_dtype(hf_dtype) is np_dtype


def test_convert_hf_features_raises_type_error():
  with pytest.raises(TypeError, match='Type <.+> is not supported.'):
    huggingface_utils.convert_hf_features('I am no features')


def test_convert_hf_features_raises_value_error():
  with pytest.raises(
      ValueError, match=r'List \[.+\] should have a length of 1.'
  ):
    huggingface_utils.convert_hf_features(
        [hf_datasets.Value('int32'), hf_datasets.Value('int32')]
    )


@pytest.mark.parametrize(
    'hf_features,tfds_features',
    [
        (
            hf_datasets.Features(
                id=hf_datasets.Value('string'),
                meta={
                    'left_context': hf_datasets.Value('string'),
                    'partial_evidence': [{
                        'start_id': hf_datasets.Value('int32'),
                        'meta': {
                            'evidence_span': [hf_datasets.Value('string')]
                        },
                    }],
                },
            ),
            feature_lib.FeaturesDict({
                'id': feature_lib.Scalar(dtype=np.str_),
                'meta': feature_lib.FeaturesDict({
                    'left_context': feature_lib.Scalar(dtype=np.str_),
                    'partial_evidence': feature_lib.Sequence({
                        'meta': feature_lib.FeaturesDict({
                            'evidence_span': feature_lib.Sequence(
                                feature_lib.Scalar(dtype=np.str_)
                            ),
                        }),
                        'start_id': feature_lib.Scalar(dtype=np.int32),
                    }),
                }),
            }),
        ),
        (
            hf_datasets.Audio(sampling_rate=48000),
            feature_lib.Audio(sample_rate=48000, dtype=np.int32),
        ),
    ],
)
def test_convert_hf_features(hf_features, tfds_features):
  assert repr(huggingface_utils.convert_hf_features(hf_features)) == repr(
      tfds_features
  )


@pytest.fixture(name='mock_list_datasets')
def _list_datasets(monkeypatch):
  def mock_list_datasets():
    return [
        huggingface_hub.DatasetInfo(id=name)
        for name in ('mnist', 'bigscience/P3', 'x', 'x/Y-z', 'fashion_mnist')
    ]

  monkeypatch.setattr(huggingface_hub, 'list_datasets', mock_list_datasets)


def test_to_huggingface_name_raises(mock_list_datasets):
  del mock_list_datasets
  with pytest.raises(
      registered.DatasetNotFoundError,
      match='"z" is not listed in Huggingface datasets.',
  ):
    assert huggingface_utils.to_huggingface_name('z')


@pytest.mark.parametrize(
    'tfds_dataset_name,hf_dataset_name',
    [
        ('x', 'x'),
        ('X', 'x'),
        ('bigscience__p3', 'bigscience/P3'),
        ('fashion_mnist', 'fashion_mnist'),
        ('x__y_z', 'x/Y-z'),
    ],
)
def test_to_huggingface_name(
    mock_list_datasets, tfds_dataset_name, hf_dataset_name
):
  del mock_list_datasets
  assert (
      huggingface_utils.to_huggingface_name(tfds_dataset_name)
      == hf_dataset_name
  )



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: smart_control_reward.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1asmart_control_reward.proto\x12#smart_buildings.smart_control.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\xe9\t\n\nRewardInfo\x12\x33\n\x0fstart_timestamp\x18\x01'
    b' \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\rend_timestamp\x18\x02'
    b' \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08\x61gent_id\x18\x03'
    b' \x01(\t\x12\x13\n\x0bscenario_id\x18\x04'
    b' \x01(\t\x12_\n\x11zone_reward_infos\x18\x05'
    b' \x03(\x0b\x32\x44.smart_buildings.smart_control.proto.RewardInfo.ZoneRewardInfosEntry\x12l\n\x18\x61ir_handler_reward_infos\x18\x06'
    b' \x03(\x0b\x32J.smart_buildings.smart_control.proto.RewardInfo.AirHandlerRewardInfosEntry\x12\x63\n\x13\x62oiler_reward_infos\x18\x07'
    b' \x03(\x0b\x32\x46.smart_buildings.smart_control.proto.RewardInfo.BoilerRewardInfosEntry\x1a\xcc\x01\n\x0eZoneRewardInfo\x12$\n\x1cheating_setpoint_temperature\x18\x01'
    b' \x01(\x02\x12$\n\x1c\x63ooling_setpoint_temperature\x18\x02'
    b' \x01(\x02\x12\x1c\n\x14zone_air_temperature\x18\x03'
    b' \x01(\x02\x12\x1e\n\x16\x61ir_flow_rate_setpoint\x18\x04'
    b' \x01(\x02\x12\x15\n\rair_flow_rate\x18\x05'
    b' \x01(\x02\x12\x19\n\x11\x61verage_occupancy\x18\x06'
    b' \x01(\x02\x1an\n\x14\x41irHandlerRewardInfo\x12%\n\x1d\x62lower_electrical_energy_rate\x18\x01'
    b" \x01(\x02\x12/\n'air_conditioning_electrical_energy_rate\x18\x02"
    b" \x01(\x02\x1a`\n\x10\x42oilerRewardInfo\x12'\n\x1fnatural_gas_heating_energy_rate\x18\x01"
    b' \x01(\x02\x12#\n\x1bpump_electrical_energy_rate\x18\x02'
    b' \x01(\x02\x1av\n\x14ZoneRewardInfosEntry\x12\x0b\n\x03key\x18\x01'
    b' \x01(\t\x12M\n\x05value\x18\x02'
    b' \x01(\x0b\x32>.smart_buildings.smart_control.proto.RewardInfo.ZoneRewardInfo:\x02\x38\x01\x1a\x82\x01\n\x1a\x41irHandlerRewardInfosEntry\x12\x0b\n\x03key\x18\x01'
    b' \x01(\t\x12S\n\x05value\x18\x02'
    b' \x01(\x0b\x32\x44.smart_buildings.smart_control.proto.RewardInfo.AirHandlerRewardInfo:\x02\x38\x01\x1az\n\x16\x42oilerRewardInfosEntry\x12\x0b\n\x03key\x18\x01'
    b' \x01(\t\x12O\n\x05value\x18\x02'
    b' \x01(\x0b\x32@.smart_buildings.smart_control.proto.RewardInfo.BoilerRewardInfo:\x02\x38\x01"\xe4\x04\n\x0eRewardResponse\x12\x1a\n\x12\x61gent_reward_value\x18\x01'
    b' \x01(\x02\x12\x1b\n\x13productivity_reward\x18\x02'
    b' \x01(\x02\x12\x1f\n\x17\x65lectricity_energy_cost\x18\x03'
    b' \x01(\x02\x12\x1f\n\x17natural_gas_energy_cost\x18\x04'
    b' \x01(\x02\x12\x16\n\x0e\x63\x61rbon_emitted\x18\x05'
    b' \x01(\x02\x12\x13\n\x0b\x63\x61rbon_cost\x18\x06'
    b' \x01(\x02\x12\x1b\n\x13productivity_weight\x18\x07'
    b' \x01(\x02\x12\x1a\n\x12\x65nergy_cost_weight\x18\x08'
    b' \x01(\x02\x12\x1e\n\x16\x63\x61rbon_emission_weight\x18\t'
    b' \x01(\x02\x12\x1b\n\x13person_productivity\x18\n'
    b' \x01(\x02\x12\x17\n\x0ftotal_occupancy\x18\x0b'
    b' \x01(\x02\x12\x14\n\x0creward_scale\x18\x0c'
    b' \x01(\x02\x12\x14\n\x0creward_shift\x18\r'
    b' \x01(\x02\x12\x1b\n\x13productivity_regret\x18\x0e'
    b' \x01(\x02\x12&\n\x1enormalized_productivity_regret\x18\x0f'
    b' \x01(\x02\x12\x1e\n\x16normalized_energy_cost\x18\x10'
    b' \x01(\x02\x12"\n\x1anormalized_carbon_emission\x18\x11'
    b' \x01(\x02\x12\x33\n\x0fstart_timestamp\x18\x12'
    b' \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\rend_timestamp\x18\x13'
    b' \x01(\x0b\x32\x1a.google.protobuf.Timestampb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, 'smart_control_reward_pb2', globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _REWARDINFO_ZONEREWARDINFOSENTRY._options = None
  _REWARDINFO_ZONEREWARDINFOSENTRY._serialized_options = b'8\001'
  _REWARDINFO_AIRHANDLERREWARDINFOSENTRY._options = None
  _REWARDINFO_AIRHANDLERREWARDINFOSENTRY._serialized_options = b'8\001'
  _REWARDINFO_BOILERREWARDINFOSENTRY._options = None
  _REWARDINFO_BOILERREWARDINFOSENTRY._serialized_options = b'8\001'
  _REWARDINFO._serialized_start = 101
  _REWARDINFO._serialized_end = 1358
  _REWARDINFO_ZONEREWARDINFO._serialized_start = 567
  _REWARDINFO_ZONEREWARDINFO._serialized_end = 771
  _REWARDINFO_AIRHANDLERREWARDINFO._serialized_start = 773
  _REWARDINFO_AIRHANDLERREWARDINFO._serialized_end = 883
  _REWARDINFO_BOILERREWARDINFO._serialized_start = 885
  _REWARDINFO_BOILERREWARDINFO._serialized_end = 981
  _REWARDINFO_ZONEREWARDINFOSENTRY._serialized_start = 983
  _REWARDINFO_ZONEREWARDINFOSENTRY._serialized_end = 1101
  _REWARDINFO_AIRHANDLERREWARDINFOSENTRY._serialized_start = 1104
  _REWARDINFO_AIRHANDLERREWARDINFOSENTRY._serialized_end = 1234
  _REWARDINFO_BOILERREWARDINFOSENTRY._serialized_start = 1236
  _REWARDINFO_BOILERREWARDINFOSENTRY._serialized_end = 1358
  _REWARDRESPONSE._serialized_start = 1361
  _REWARDRESPONSE._serialized_end = 1973
# @@protoc_insertion_point(module_scope)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pathlib
import re
import tempfile
import time
from typing import Union
from unittest import mock

from etils import epath
import numpy as np
import pytest
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import features
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import read_only_builder
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.proto import dataset_info_pb2
from tensorflow_datasets.core.proto import feature_pb2
from tensorflow_datasets.image_classification import mnist
from tensorflow_datasets.testing.dummy_config_based_datasets.dummy_ds_1 import dummy_ds_1_dataset_builder

from google.protobuf import text_format


_TFDS_DIR = utils.tfds_path()
_DATA_DIR = os.path.join(_TFDS_DIR, "testing", "test_data", "dataset_info")
_INFO_DIR = os.path.join(_DATA_DIR, "mnist", "3.0.1")
_INFO_DIR_UNLABELED = os.path.join(_DATA_DIR, "mnist_unlabeled", "3.0.1")
_NON_EXISTENT_DIR = os.path.join(_TFDS_DIR, "non_existent_dir")

DummyDatasetSharedGenerator = testing.DummyDatasetSharedGenerator


class RandomShapedImageGenerator(DummyDatasetSharedGenerator):

  BUILDER_CONFIGS = [
      dataset_builder.BuilderConfig(
          name="random_shaped_image_config",
      )
  ]

  def _info(self):
    return dataset_info.DatasetInfo(
        builder=self,
        features=features.FeaturesDict({"im": features.Image()}),
        supervised_keys=("im", "im"),
        metadata=dataset_info.MetadataDict(),
    )

  def _generate_examples(self, range_):
    self.info.metadata["some_key"] = 123

    for i in range_:
      height = np.random.randint(5, high=10)
      width = np.random.randint(5, high=10)
      yield i, {
          "im": np.random.randint(
              0, 255, size=(height, width, 3), dtype=np.uint8
          )
      }


class DatasetInfoTest(testing.TestCase):

  @classmethod
  def setUpClass(cls):
    super(DatasetInfoTest, cls).setUpClass()
    cls._tfds_tmp_dir = testing.make_tmp_dir()
    cls._builder = mnist.MNIST(data_dir=cls._tfds_tmp_dir)

  @classmethod
  def tearDownClass(cls):
    super(DatasetInfoTest, cls).tearDownClass()
    testing.rm_tmp_dir(cls._tfds_tmp_dir)

  def test_non_existent_dir(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    with self.assertRaisesWithPredicateMatch(
        dataset_info.DatasetInfoFileError,
        "Could not read dataset info from",
    ):
      info.read_from_directory(_NON_EXISTENT_DIR)

  def test_reading_different_version(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info._identity.version = utils.Version("2.0.0")
    with pytest.raises(
        AssertionError,
        match=(
            "The constructed DatasetInfo instance and the restored proto"
            " version do not match"
        ),
    ):
      # The dataset in _INFO_DIR has version 3.0.1 whereas the builder is 2.0.0
      info.read_from_directory(_INFO_DIR)

  def test_reading(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info.read_from_directory(_INFO_DIR)

    # Assert that we read the file and initialized DatasetInfo.
    self.assertTrue(info.initialized)
    self.assertEqual("mnist", info.name)
    self.assertEqual("mnist/3.0.1", info.full_name)

    # Test splits are initialized properly.
    split_dict = info.splits

    # Assert they are the correct number.
    self.assertLen(split_dict, 2)

    # Assert on what they are
    self.assertIn("train", split_dict)
    self.assertIn("test", split_dict)

    # Assert that this is computed correctly.
    self.assertEqual(40, info.splits.total_num_examples)
    self.assertEqual(11594722, info.dataset_size)

    self.assertEqual("image", info.supervised_keys[0])
    self.assertEqual("label", info.supervised_keys[1])
    self.assertEqual(
        info.module_name, "tensorflow_datasets.image_classification.mnist"
    )
    self.assertEqual(False, info.disable_shuffling)

    self.assertEqual(info.version, utils.Version("3.0.1"))
    self.assertEqual(info.release_notes, {})

  def test_disable_shuffling(self):
    info = dataset_info.DatasetInfo(
        builder=self._builder, disable_shuffling=True
    )
    info.read_from_directory(_INFO_DIR)

    self.assertEqual(True, info.disable_shuffling)

  def test_reading_empty_properties(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info.read_from_directory(_INFO_DIR_UNLABELED)

    # Assert supervised_keys has not been set
    self.assertIsNone(None, info.supervised_keys)

  def test_writing(self):
    # First read in stuff.
    mnist_builder = mnist.MNIST(data_dir=_DATA_DIR)

    info = dataset_info.DatasetInfo(
        builder=mnist_builder, features=mnist_builder.info.features
    )
    info.read_from_directory(_INFO_DIR)

    # Read the json file into a string.
    with tf.io.gfile.GFile(dataset_info.dataset_info_path(_INFO_DIR)) as f:
      existing_json = json.load(f)

    # Now write to a temp directory.
    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      info.write_to_directory(tmp_dir)

      # Read the newly written json file into a string.
      with tf.io.gfile.GFile(dataset_info.dataset_info_path(tmp_dir)) as f:
        new_json = json.load(f)

      # Read the newly written LICENSE file into a string.
      with tf.io.gfile.GFile(dataset_info.license_path(tmp_dir)) as f:
        license_ = f.read()

    # Assert what was read and then written and read again is the same.
    self.assertEqual(existing_json, new_json)

    # Assert correct license was written.
    self.assertEqual(existing_json["redistributionInfo"]["license"], license_)

    # Do not check the full string as it display the generated path.
    self.assertEqual(_INFO_STR % mnist_builder.data_dir, repr(info))
    self.assertIn("'test': <SplitInfo num_examples=", repr(info))

  def test_restore_after_modification(self):
    # Create a DatasetInfo
    info = dataset_info.DatasetInfo(
        builder=self._builder,
        description="A description",
        supervised_keys=("input", "output"),
        homepage="http://some-location",
        citation="some citation",
        license="some license",
    )
    info.as_proto.config_tags.extend(["foo", "bar"])
    info.download_size = 456
    filepath_template = "{DATASET}-{SPLIT}.{FILEFORMAT}-{SHARD_X_OF_Y}"
    info.as_proto.splits.add(
        name="train", num_bytes=512, filepath_template=filepath_template
    )
    info.as_proto.splits.add(
        name="validation", num_bytes=64, filepath_template=filepath_template
    )
    info.as_proto.schema.feature.add()
    info.as_proto.schema.feature.add()  # Add dynamic statistics
    info.download_checksums = {
        "url1": "some checksum",
        "url2": "some other checksum",
    }

    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      # Save it
      info.write_to_directory(tmp_dir)

      # If fields are not defined, then everything is restored from disk.
      restored_info = dataset_info.DatasetInfo(builder=self._builder)
      restored_info.read_from_directory(tmp_dir)
      self.assertProtoEquals(info.as_proto, restored_info.as_proto)

    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      # Save it
      info.write_to_directory(tmp_dir)

      # If fields are defined, then the code version is kept
      restored_info = dataset_info.DatasetInfo(
          builder=self._builder,
          supervised_keys=("input (new)", "output (new)"),
          homepage="http://some-location-new",
          citation="some citation (new)",
          redistribution_info={"license": "some license (new)"},
      )
      restored_info.download_size = 789
      restored_info.as_proto.splits.add(name="validation", num_bytes=288)
      restored_info.as_proto.schema.feature.add()
      restored_info.as_proto.schema.feature.add()
      restored_info.as_proto.schema.feature.add()
      restored_info.as_proto.schema.feature.add()  # Add dynamic statistics
      restored_info.download_checksums = {
          "url2": "some other checksum (new)",
          "url3": "some checksum (new)",
      }

      restored_info.read_from_directory(tmp_dir)

      # Even though restored_info has been restored, informations defined in
      # the code overwrite informations from the json file.
      self.assertEqual(restored_info.description, "A description")
      self.assertEqual(restored_info.version, utils.Version("3.0.1"))
      self.assertEqual(restored_info.release_notes, {})
      self.assertEqual(
          restored_info.supervised_keys, ("input (new)", "output (new)")
      )
      self.assertEqual(restored_info.homepage, "http://some-location-new")
      self.assertEqual(restored_info.citation, "some citation (new)")
      self.assertEqual(
          restored_info.redistribution_info.license, "some license (new)"
      )
      self.assertEqual(restored_info.download_size, 789)
      self.assertEqual(restored_info.dataset_size, 576)
      self.assertLen(restored_info.as_proto.schema.feature, 4)
      self.assertEqual(
          restored_info.download_checksums,
          {
              "url2": "some other checksum (new)",
              "url3": "some checksum (new)",
          },
      )

  def test_reading_from_gcs_bucket(self):
    # The base TestCase prevents GCS access, so we explicitly ask it to restore
    # access here.
    with testing.enable_gcs_access():
      mnist_builder = mnist.MNIST(
          data_dir=tempfile.mkdtemp(dir=self.get_temp_dir())
      )
      info = dataset_info.DatasetInfo(builder=mnist_builder)
      info = mnist_builder.info

      # A nominal check to see if we read it.
      self.assertTrue(info.initialized)
      self.assertEqual(10000, info.splits["test"].num_examples)

  def test_str_smoke(self):
    info = mnist.MNIST(data_dir="/tmp/some_dummy_dir").info
    _ = str(info)

  def test_metadata(self):
    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      builder = RandomShapedImageGenerator(data_dir=tmp_dir)
      builder.download_and_prepare()
      # Metadata should have been created
      self.assertEqual(builder.info.metadata, {"some_key": 123})

      # Metadata should have been restored
      builder2 = RandomShapedImageGenerator(data_dir=tmp_dir)
      self.assertEqual(builder2.info.metadata, {"some_key": 123})

      # Metadata should have been restored even if the builder code was not
      # available and we restored from files.
      builder3 = read_only_builder.builder_from_files(
          builder.name,
          data_dir=tmp_dir,
      )
      self.assertEqual(builder3.info.metadata, {"some_key": 123})

      # Test whether the metadata is copyable.
      builder = RandomShapedImageGenerator(data_dir=tmp_dir)
      self.assertIsInstance(
          builder.info.metadata, dataset_info.LazyMetadataDict
      )
      metadata_copy = {**builder.info.metadata}
      self.assertEqual(metadata_copy, {"some_key": 123})

  def test_redistribution_info(self):
    info = dataset_info.DatasetInfo(
        builder=self._builder, license="some license"
    )
    info_proto = info.as_proto
    self.assertEqual(info_proto.redistribution_info.license, "some license")

  def test_updates_on_bucket_info(self):
    info = dataset_info.DatasetInfo(
        builder=self._builder, description="won't be updated"
    )
    # No statistics in the above.
    self.assertEqual(0, info.splits.total_num_examples)
    self.assertEmpty(info.as_proto.schema.feature)

    # Partial update will happen here.
    info.read_from_directory(_INFO_DIR)

    # Assert that description (things specified in the code) didn't change
    # but statistics are updated.
    self.assertEqual("won't be updated", info.description)

    # These are dynamically computed, so will be updated.
    self.assertEqual(40, info.splits.total_num_examples)
    self.assertLen(info.as_proto.schema.feature, 2)

  def test_set_splits_normal(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    split_info1 = splits_lib.SplitInfo(
        name="train", shard_lengths=[1, 2], num_bytes=0
    )
    split_info2 = splits_lib.SplitInfo(
        name="test", shard_lengths=[1], num_bytes=0
    )
    split_dict = splits_lib.SplitDict(split_infos=[split_info1, split_info2])
    info.set_splits(split_dict)
    self.assertEqual(str(info.splits), str(split_dict))
    self.assertEqual(
        str(info.as_proto.splits),
        str([split_info1.to_proto(), split_info2.to_proto()]),
    )

  def test_set_splits_incorrect_dataset_name(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    split_info1 = splits_lib.SplitInfo(
        name="train",
        shard_lengths=[1, 2],
        num_bytes=0,
        filename_template=naming.ShardedFileTemplate(
            dataset_name="some_other_dataset",
            split="train",
            data_dir=info.data_dir,
            filetype_suffix="tfrecord",
        ),
    )
    split_dict = splits_lib.SplitDict(split_infos=[split_info1])
    with pytest.raises(
        AssertionError, match="SplitDict contains SplitInfo for split"
    ):
      info.set_splits(split_dict)

  def test_set_splits_multi_split_info(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    split_info1 = splits_lib.SplitInfo(
        name="train", shard_lengths=[1, 2], num_bytes=0
    )
    split_info2 = splits_lib.SplitInfo(
        name="test", shard_lengths=[1], num_bytes=0
    )
    multi_split_info1 = splits_lib.MultiSplitInfo(
        name="train", split_infos=[split_info1]
    )
    multi_split_info2 = splits_lib.MultiSplitInfo(
        name="test", split_infos=[split_info2]
    )
    split_dict = splits_lib.SplitDict(
        split_infos=[multi_split_info1, multi_split_info2]
    )
    info.set_splits(split_dict)
    self.assertEqual(str(info.splits), str(split_dict))
    self.assertEqual(
        str(info.as_proto.splits),
        str([split_info1.to_proto(), split_info2.to_proto()]),
    )

  def test_set_file_format_override_fails(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info.set_file_format(file_adapters.FileFormat.TFRECORD)
    self.assertEqual(info.file_format, file_adapters.FileFormat.TFRECORD)
    with pytest.raises(
        ValueError,
        match=(
            "File format is already set to FileFormat.TFRECORD. Got"
            " FileFormat.RIEGELI"
        ),
    ):
      info.set_file_format(file_adapters.FileFormat.RIEGELI)

  def test_set_file_format_override(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info.set_file_format(file_adapters.FileFormat.TFRECORD)
    self.assertEqual(info.file_format, file_adapters.FileFormat.TFRECORD)
    info.set_file_format(file_adapters.FileFormat.RIEGELI, override=True)
    self.assertEqual(info.file_format, file_adapters.FileFormat.RIEGELI)

  def test_set_file_format_override_failes_when_fully_initialized(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info.set_file_format(file_adapters.FileFormat.TFRECORD)
    info._fully_initialized = True
    self.assertEqual(info.file_format, file_adapters.FileFormat.TFRECORD)
    with pytest.raises(
        ValueError,
        match=(
            "File format is already set to FileFormat.TFRECORD. Got"
            " FileFormat.RIEGELI"
        ),
    ):
      info.set_file_format(file_adapters.FileFormat.RIEGELI)

  def test_set_file_format_override_fully_initialized(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info.set_file_format(file_adapters.FileFormat.TFRECORD)
    info._fully_initialized = True
    self.assertEqual(info.file_format, file_adapters.FileFormat.TFRECORD)
    info.set_file_format(
        file_adapters.FileFormat.RIEGELI,
        override=True,
        override_if_initialized=True,
    )
    self.assertEqual(info.file_format, file_adapters.FileFormat.RIEGELI)

  def test_update_info_proto_with_features(self):
    info_proto = dataset_info.DatasetInfo(builder=self._builder).as_proto
    new_features = features.FeaturesDict({"text": features.Text()})
    new_info = dataset_info.update_info_proto_with_features(
        info_proto, new_features
    )
    self.assertEqual(new_info.features, new_features.to_proto())


@pytest.mark.parametrize(
    "file_format",
    [
        file_adapters.FileFormat.TFRECORD,
    ],
)
def test_file_format_save_restore(
    tmp_path: pathlib.Path,
    file_format: file_adapters.FileFormat,
):
  builder = testing.DummyDataset(data_dir=tmp_path, file_format=file_format)

  assert isinstance(builder.info.file_format, file_adapters.FileFormat)
  assert builder.info.file_format is file_format

  builder.download_and_prepare()

  # When restoring the builder, we do not provide the `file_format=`
  # yet it is correctly restored
  builder2 = testing.DummyDataset(data_dir=tmp_path)
  assert builder2.info.file_format is file_format

  # Explicitly passing the correct format is accepted.
  builder3 = testing.DummyDataset(data_dir=tmp_path, file_format=file_format)
  assert builder3.info.file_format is file_format

  # Providing an inconsistent format is rejected.
  with pytest.raises(ValueError, match="File format is already set to"):
    different_file_format = {
        file_adapters.FileFormat.TFRECORD: file_adapters.FileFormat.RIEGELI,
        file_adapters.FileFormat.RIEGELI: file_adapters.FileFormat.TFRECORD,
    }[file_format]
    testing.DummyDataset(data_dir=tmp_path, file_format=different_file_format)


def test_file_format_values(tmp_path: pathlib.Path):
  # Default file format
  builder = testing.DummyDataset(data_dir=tmp_path, file_format=None)
  assert builder.info.file_format == file_adapters.FileFormat.TFRECORD

  # str accepted
  builder = testing.DummyDataset(data_dir=tmp_path, file_format="riegeli")
  assert builder.info.file_format == file_adapters.FileFormat.RIEGELI

  # file_adapters.FileFormat accepted
  builder = testing.DummyDataset(
      data_dir=tmp_path, file_format=file_adapters.FileFormat.RIEGELI
  )
  assert builder.info.file_format == file_adapters.FileFormat.RIEGELI

  # Unknown value
  with pytest.raises(ValueError, match="is not a valid FileFormat"):
    testing.DummyDataset(data_dir=tmp_path, file_format="arrow")


@pytest.fixture(scope="session")
def generator_builder():
  return RandomShapedImageGenerator(data_dir=testing.make_tmp_dir())


def test_dataset_info_from_proto(generator_builder):  # pylint: disable=redefined-outer-name
  train = dataset_info_pb2.SplitInfo(
      name="train", num_shards=2, shard_lengths=[4, 5]
  )
  test = dataset_info_pb2.SplitInfo(
      name="test", num_shards=3, shard_lengths=[1, 2, 3]
  )
  text_feature = feature_pb2.Feature(
      python_class_name="tensorflow_datasets.core.features.text_feature.Text",
      text=feature_pb2.TextFeature(),
  )
  proto = dataset_info_pb2.DatasetInfo(
      name="random_shaped_image_generator",
      config_name=generator_builder.builder_config.name,
      version=str(generator_builder.version),
      features=feature_pb2.Feature(
          python_class_name=(
              "tensorflow_datasets.core.features.features_dict.FeaturesDict"
          ),
          features_dict=feature_pb2.FeaturesDict(
              features={"text": text_feature}
          ),
      ),
      splits=[train, test],
  )
  result = dataset_info.DatasetInfo.from_proto(
      builder=generator_builder,
      proto=proto,
  )
  assert result.splits["test"].shard_lengths == test.shard_lengths
  assert result.splits["train"].shard_lengths == train.shard_lengths
  assert set(result.features.keys()) == {"text"}
  assert result.version == generator_builder.version


def test_supervised_keys_from_proto():
  proto = text_format.Parse(
      text="""
  tuple: {
      items: [
        {
          dict: {
            dict: {
              key: "f2"
              value: { feature_key: "f2" }
            },
            dict: {
              key: "f1"
              value: { feature_key: "f1" }
            },
          }
        },
        {
          feature_key: "target"
        }
      ]
    }
  """,
      message=dataset_info_pb2.SupervisedKeys(),
  )
  supervised_keys = dataset_info._supervised_keys_from_proto(proto=proto)
  assert str(supervised_keys) == "({'f1': 'f1', 'f2': 'f2'}, 'target')"


def test_supervised_keys_from_proto_different_ordering():
  proto1 = text_format.Parse(
      text="""
  tuple: {
      items: [
        {
          dict: {
            dict: {
              key: "f1"
              value: { feature_key: "f1" }
            },
            dict: {
              key: "f2"
              value: { feature_key: "f2" }
            },
            dict: {
              key: "f3"
              value: { feature_key: "f3" }
            },
          }
        },
        {
          feature_key: "target"
        }
      ]
    }
  """,
      message=dataset_info_pb2.SupervisedKeys(),
  )
  proto2 = text_format.Parse(
      text="""
  tuple: {
      items: [
        {
          dict: {
            dict: {
              key: "f3"
              value: { feature_key: "f3" }
            },
            dict: {
              key: "f2"
              value: { feature_key: "f2" }
            },
            dict: {
              key: "f1"
              value: { feature_key: "f1" }
            },
          }
        },
        {
          feature_key: "target"
        }
      ]
    }
  """,
      message=dataset_info_pb2.SupervisedKeys(),
  )
  supervised_keys1 = dataset_info._supervised_keys_from_proto(proto=proto1)
  supervised_keys2 = dataset_info._supervised_keys_from_proto(proto=proto2)
  assert str(supervised_keys1) == str(supervised_keys2)


@pytest.mark.parametrize(
    ("license_", "redistribution_info", "expected_license"),
    (
        ("a", None, "a"),
        (None, {"license": "a"}, "a"),
        ("a", {"license": "a"}, "a"),
        (None, None, None),
    ),
)
def test_create_redistribution_info_proto(
    license_, redistribution_info, expected_license
):
  result = dataset_info._create_redistribution_info_proto(
      license=license_, redistribution_info=redistribution_info
  )
  if expected_license is None:
    assert result is None
  else:
    assert result.license == expected_license


def test_create_redistribution_info_proto_inconsistent():
  with pytest.raises(
      ValueError, match="License specified twice and inconsistently"
  ):
    dataset_info._create_redistribution_info_proto(
        license="a", redistribution_info={"license": "b"}
    )


def test_create_redistribution_info_proto_unsupported_fields():
  with pytest.raises(
      ValueError, match="`redistribution_info` contains unsupported keys"
  ):
    dataset_info._create_redistribution_info_proto(
        license="a", redistribution_info={"unsupported": "b"}
    )


def test_supports_file_format():
  dataset_info_proto = dataset_info_pb2.DatasetInfo(
      file_format=file_adapters.FileFormat.TFRECORD.value,
      alternative_file_formats=[file_adapters.FileFormat.RIEGELI.value],
  )
  assert dataset_info.supports_file_format(
      dataset_info_proto, file_format=file_adapters.FileFormat.TFRECORD
  )
  assert dataset_info.supports_file_format(
      dataset_info_proto, file_format=file_adapters.FileFormat.RIEGELI
  )
  assert not dataset_info.supports_file_format(
      dataset_info_proto, file_format=file_adapters.FileFormat.PARQUET
  )


class GetSplitInfoFromProtoTest(testing.TestCase):

  def _dataset_info_proto_with_splits(self):
    return dataset_info_pb2.DatasetInfo(
        name="dataset",
        file_format="tfrecord",
        alternative_file_formats=["riegeli"],
        splits=[
            dataset_info_pb2.SplitInfo(
                name="train",
                shard_lengths=[1, 2, 3],
                num_bytes=42,
            ),
            dataset_info_pb2.SplitInfo(
                name="test",
                shard_lengths=[1, 2, 3],
                num_bytes=42,
                filepath_template="{SPLIT}.{FILEFORMAT}-{SHARD_X_OF_Y}",
            ),
        ],
    )

  def test_get_split_dict_from_proto(self):
    actual = dataset_info.get_split_dict_from_proto(
        dataset_info_proto=self._dataset_info_proto_with_splits(),
        data_dir="/data",
        file_format=file_adapters.FileFormat.PARQUET,
    )
    assert set(["train", "test"]) == set(actual.keys())

    train = actual["train"]
    assert train.name == "train"
    assert train.shard_lengths == [1, 2, 3]
    assert train.num_bytes == 42
    assert train.filename_template.dataset_name == "dataset"
    assert train.filename_template.template == naming.DEFAULT_FILENAME_TEMPLATE
    assert train.filename_template.filetype_suffix == "parquet"

    test = actual["test"]
    assert test.name == "test"
    assert test.shard_lengths == [1, 2, 3]
    assert test.num_bytes == 42
    assert test.filename_template.dataset_name == "dataset"
    assert (
        test.filename_template.template == "{SPLIT}.{FILEFORMAT}-{SHARD_X_OF_Y}"
    )
    assert test.filename_template.filetype_suffix == "parquet"

  def test_get_split_info_from_proto_undefined_filename_template(self):
    actual = dataset_info.get_split_info_from_proto(
        dataset_info_proto=self._dataset_info_proto_with_splits(),
        split_name="train",
        data_dir="/path/to/data",
        file_format=file_adapters.FileFormat.TFRECORD,
    )
    assert actual.name == "train"
    assert actual.shard_lengths == [1, 2, 3]
    assert actual.num_bytes == 42
    assert actual.filename_template.dataset_name == "dataset"
    assert actual.filename_template.template == naming.DEFAULT_FILENAME_TEMPLATE

  def test_get_split_info_from_proto_defined_filename_template(self):
    actual = dataset_info.get_split_info_from_proto(
        dataset_info_proto=self._dataset_info_proto_with_splits(),
        split_name="test",
        data_dir="/path/to/data",
        file_format=file_adapters.FileFormat.TFRECORD,
    )
    assert actual.name == "test"
    assert actual.shard_lengths == [1, 2, 3]
    assert actual.filename_template.dataset_name == "dataset"
    assert (
        actual.filename_template.template
        == "{SPLIT}.{FILEFORMAT}-{SHARD_X_OF_Y}"
    )

  def test_get_split_info_from_proto_non_existing_split(self):
    actual = dataset_info.get_split_info_from_proto(
        dataset_info_proto=self._dataset_info_proto_with_splits(),
        split_name="undefined",
        data_dir="/path/to/data",
        file_format=file_adapters.FileFormat.TFRECORD,
    )
    assert actual is None

  def test_get_split_info_from_proto_unavailable_format(self):
    with pytest.raises(
        ValueError,
        match=re.escape(
            "File format parquet does not match available dataset file formats:"
            " ['riegeli', 'tfrecord']."
        ),
    ):
      dataset_info.get_split_info_from_proto(
          dataset_info_proto=self._dataset_info_proto_with_splits(),
          split_name="undefined",
          data_dir="/path/to/data",
          file_format=file_adapters.FileFormat.PARQUET,
      )


# pylint: disable=g-inconsistent-quotes
_INFO_STR = '''tfds.core.DatasetInfo(
    name='mnist',
    full_name='mnist/3.0.1',
    description="""
    The MNIST database of handwritten digits.
    """,
    homepage='https://storage.googleapis.com/cvdf-datasets/mnist/',
    data_dir='%s',
    file_format=tfrecord,
    download_size=1.95 KiB,
    dataset_size=11.06 MiB,
    features=FeaturesDict({
        'image': Image(shape=(28, 28, 1), dtype=uint8),
        'label': ClassLabel(shape=(), dtype=int64, num_classes=10),
    }),
    supervised_keys=('image', 'label'),
    disable_shuffling=False,
    nondeterministic_order=False,
    splits={
        'test': <SplitInfo num_examples=20, num_shards=1>,
        'train': <SplitInfo num_examples=20, num_shards=1>,
    },
    citation="""@article{lecun2010mnist,
      title={MNIST handwritten digit database},
      author={LeCun, Yann and Cortes, Corinna and Burges, CJ},
      journal={ATT Labs [Online]. Available: http://yann. lecun. com/exdb/mnist},
      volume={2},
      year={2010}
    }
    """,
    redistribution_info=license: "test license",
)'''
# pylint: enable=g-inconsistent-quotes


class LazyMetadataDictTest(testing.TestCase):

  def setUp(self):
    super().setUp()
    self.data_dir = epath.Path(self.tmp_dir)
    metadata_file = self.data_dir / "metadata.json"
    metadata_file.write_text(json.dumps({"test": "test123"}))

  def test_load_metadata(self):
    metadata_dict = dataset_info.LazyMetadataDict(self.data_dir)
    self.assertEqual(metadata_dict["test"], "test123")

  @mock.patch.object(dataset_info, "_load_metadata_from_file")
  def test_get_item(self, mock_load_metadata_from_file):
    mock_load_metadata_from_file.return_value = {"test": "test123"}

    metadata_dict = dataset_info.LazyMetadataDict(self.data_dir)
    mock_load_metadata_from_file.assert_not_called()

    self.assertEqual(metadata_dict["test"], "test123")
    mock_load_metadata_from_file.assert_called_with(self.data_dir)

    self.assertEqual(sorted(metadata_dict.keys()), ["test"])

    # Update the metadata.
    metadata_dict["abc"] = "def"
    self.assertEqual(metadata_dict["abc"], "def")

  @mock.patch.object(dataset_info, "_load_metadata_from_file")
  def test_keys(self, mock_load_metadata_from_file):
    mock_load_metadata_from_file.return_value = {"test": "test123"}
    metadata_dict = dataset_info.LazyMetadataDict(self.data_dir)
    mock_load_metadata_from_file.assert_not_called()
    actual_keys = sorted(metadata_dict.keys())
    mock_load_metadata_from_file.assert_called_with(self.data_dir)
    self.assertEqual(actual_keys, ["test"])

  @mock.patch.object(dataset_info, "_load_metadata_from_file")
  def test_items(self, mock_load_metadata_from_file):
    mock_load_metadata_from_file.return_value = {"test": "test123"}
    metadata_dict = dataset_info.LazyMetadataDict(self.data_dir)
    mock_load_metadata_from_file.assert_not_called()
    self.assertNotEmpty(metadata_dict.items())
    mock_load_metadata_from_file.assert_called_with(self.data_dir)

  @mock.patch.object(dataset_info, "_load_metadata_from_file")
  def test_updating_before_loading(self, mock_load_metadata_from_file):
    metadata_dict = dataset_info.LazyMetadataDict(self.data_dir)
    mock_load_metadata_from_file.assert_not_called()

    metadata_dict.update({"abc": "def", "ghi": "jkl"})
    mock_load_metadata_from_file.assert_not_called()

  def test_copy(self):
    metadata_dict = dataset_info.LazyMetadataDict(self.data_dir)
    metadata_dict_copy = metadata_dict.copy()
    self.assertEqual(metadata_dict_copy["test"], "test123")


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import re
from unittest import mock
import pytest
from tensorflow_datasets import testing
from tensorflow_datasets.core import load
from tensorflow_datasets.core import registered
from tensorflow_datasets.core import splits
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.utils import py_utils
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.testing.dummy_config_based_datasets.dummy_ds_1 import dummy_ds_1_dataset_builder


class TestBuilderProvider(tfds.core.DatasetBuilderProvider):
  """Test Builder provider."""

  def __init__(
      self,
      name: str,
      dataset_builder_cls: type[tfds.core.registered.RegisteredDataset],
  ):
    self._builder_cls_cache = {name: dataset_builder_cls}
    self._available_datasets = [name]

  def has_dataset(self, name: str) -> bool:
    return name in self._available_datasets

  def get_builder_cls(
      self, name: str
  ) -> type[tfds.core.registered.RegisteredDataset]:
    if not self.has_dataset(name):
      raise ValueError(f"Dataset {name} is not available!")
    return self._builder_cls_cache[name]


class EmptyDatasetBuilder(registered.RegisteredDataset):

  def __init__(self, **kwargs):
    self.data_dir = "/some/path/to/dir"
    self.kwargs = kwargs
    self.download_called = False
    self.as_dataset_called = False

  def download_and_prepare(self, **kwargs):
    self.download_called = True
    self.download_kwargs = kwargs

  def as_dataset(self, **kwargs):
    self.as_dataset_called = True
    self.as_dataset_kwargs = kwargs
    return self

  def is_prepared(self) -> bool:
    return False

  VERSION = utils.Version("1.0.0")
  BUILDER_CONFIGS = []
  builder_configs = {}


class UnregisteredBuilder(EmptyDatasetBuilder):

  @property
  @abc.abstractmethod
  def an_abstract_property(self):
    pass


class RegisteredTest(testing.TestCase):

  def test_registered(self):
    name = "empty_dataset_builder"
    self.assertEqual(name, EmptyDatasetBuilder.name)
    self.assertIsInstance(load.builder(name), EmptyDatasetBuilder)
    self.assertIn(name, load.list_builders())

    nonexistent = "nonexistent_foobar_dataset"
    with self.assertRaisesWithPredicateMatch(
        registered.DatasetNotFoundError, "not found"
    ):
      load.builder(nonexistent)
    # Prints registered datasets
    with self.assertRaisesWithPredicateMatch(ValueError, name):
      load.builder(nonexistent)

  def test_registered_cls(self):
    name = "empty_dataset_builder"
    self.assertIs(load.builder_cls(name), EmptyDatasetBuilder)

    nonexistent = "nonexistent_foobar_dataset"
    with self.assertRaisesWithPredicateMatch(ValueError, "not found"):
      load.builder_cls(nonexistent)

    with self.assertRaisesWithPredicateMatch(
        ValueError, "`builder_cls` only accept the `dataset_name`"
    ):
      name_with_kwargs = "empty_dataset_builder/config:1.0.0"
      load.builder_cls(name_with_kwargs)

  def test_abstract(self):
    name = "unregistered_builder"
    self.assertEqual(name, UnregisteredBuilder.name)
    self.assertNotIn(name, load.list_builders())

    with self.assertRaisesWithPredicateMatch(
        TypeError, "Can't instantiate abstract class"
    ):
      load.builder(name)

  def test_builder_with_kwargs(self):
    name = "empty_dataset_builder"
    name_with_kwargs = name + "/k1=1,k2=1.,k3=foo,k4=True,k5=False"
    builder = load.builder(name_with_kwargs, data_dir="bar")
    expectations = [
        ("k1", 1),
        ("k2", 1.0),
        ("k3", "foo"),
        ("k4", True),
        ("k5", False),
    ]
    for k, v in expectations:
      self.assertEqual(type(builder.kwargs[k]), type(v))
      self.assertEqual(builder.kwargs[k], v)

  def test_builder_fullname(self):
    fullname = "empty_dataset_builder/conf1-attr:1.0.1/k1=1,k2=2"
    builder = load.builder(fullname, data_dir="bar")
    expected = {
        "k1": 1,
        "k2": 2,
        "version": "1.0.1",
        "config": "conf1-attr",
        "data_dir": "bar",
    }
    self.assertEqual(expected, builder.kwargs)

  def test_builder_camel_case(self):
    fullname = "EmptyDatasetBuilder"
    builder = load.builder(fullname)
    self.assertIsInstance(builder, EmptyDatasetBuilder)

  def test_load(self):
    name = "empty_dataset_builder/k1=1"
    data_dir = "foo"
    as_dataset_kwargs = dict(a=1, b=2)

    # EmptyDatasetBuilder returns self from as_dataset
    builder = load.load(
        name=name,
        split=splits.Split.TEST,
        data_dir=data_dir,
        download=False,
        as_dataset_kwargs=as_dataset_kwargs,
    )
    self.assertTrue(builder.as_dataset_called)
    self.assertFalse(builder.download_called)
    self.assertEqual(splits.Split.TEST, builder.as_dataset_kwargs.pop("split"))
    self.assertIsNone(builder.as_dataset_kwargs.pop("batch_size"))
    self.assertFalse(builder.as_dataset_kwargs.pop("as_supervised"))
    self.assertFalse(builder.as_dataset_kwargs.pop("decoders"))
    self.assertIsNone(builder.as_dataset_kwargs.pop("read_config"))
    self.assertFalse(builder.as_dataset_kwargs.pop("shuffle_files"))
    self.assertEqual(builder.as_dataset_kwargs, as_dataset_kwargs)
    self.assertEqual(dict(data_dir=data_dir, k1=1), builder.kwargs)

    builder = load.load(
        name,
        split=splits.Split.TRAIN,
        data_dir=data_dir,
        download=True,
        as_dataset_kwargs=as_dataset_kwargs,
    )
    self.assertTrue(builder.as_dataset_called)
    self.assertTrue(builder.download_called)

    # Tests for different batch_size
    # By default batch_size=None
    builder = load.load(name=name, split=splits.Split.TEST, data_dir=data_dir)
    self.assertIsNone(builder.as_dataset_kwargs.pop("batch_size"))
    # Setting batch_size=1
    builder = load.load(
        name=name, split=splits.Split.TEST, data_dir=data_dir, batch_size=1
    )
    self.assertEqual(1, builder.as_dataset_kwargs.pop("batch_size"))

  def test_load_all_splits(self):
    name = "empty_dataset_builder"
    # EmptyDatasetBuilder returns self from as_dataset
    builder = load.load(name=name, data_dir="foo")
    self.assertTrue(builder.as_dataset_called)
    self.assertIsNone(builder.as_dataset_kwargs.pop("split"))

  def test_load_with_config(self):
    data_dir = "foo"
    name = "empty_dataset_builder/bar/k1=1"
    # EmptyDatasetBuilder returns self from as_dataset
    builder = load.load(name=name, split=splits.Split.TEST, data_dir=data_dir)
    expected = dict(data_dir=data_dir, k1=1, config="bar")
    self.assertEqual(expected, builder.kwargs)

    name = "empty_dataset_builder/bar"
    builder = load.load(name=name, split=splits.Split.TEST, data_dir=data_dir)
    self.assertEqual(dict(data_dir=data_dir, config="bar"), builder.kwargs)

  def test_notebook_overwrite_dataset(self):
    """Redefining the same builder twice is possible on colab."""

    with mock.patch.object(py_utils, "is_notebook", lambda: True):
      name = "colab_builder"
      self.assertNotIn(name, load.list_builders())

      class ColabBuilder(EmptyDatasetBuilder):
        pass

      self.assertIn(name, load.list_builders())
      self.assertIsInstance(load.builder(name), ColabBuilder)
      old_colab_class = ColabBuilder

      class ColabBuilder(EmptyDatasetBuilder):  # pylint: disable=function-redefined
        pass

      self.assertIsInstance(load.builder(name), ColabBuilder)
      self.assertNotIsInstance(load.builder(name), old_colab_class)

  def test_duplicate_dataset(self):
    """Redefining the same builder twice should raises error."""

    class DuplicateBuilder(registered.RegisteredDataset):  # pylint: disable=unused-variable
      pass

    with self.assertRaisesWithPredicateMatch(ValueError, "already registered"):

      class DuplicateBuilder(registered.RegisteredDataset):  # pylint: disable=function-redefined
        pass

  def test_is_full_name(self):
    """Test for `is_full_name`."""
    self.assertFalse(load.is_full_name("ds/config/1.0.2/other"))
    self.assertFalse(load.is_full_name("ds/config/1.0.2/"))
    self.assertFalse(load.is_full_name("ds/config/1.2"))
    self.assertFalse(load.is_full_name("ds/config"))
    self.assertFalse(load.is_full_name("ds/1.2.*"))

    self.assertTrue(load.is_full_name("ds/config/1.0.2"))
    self.assertTrue(load.is_full_name("ds/1.0.2"))
    self.assertTrue(load.is_full_name("ds_with_number123/1.0.2"))

  def test_add_dataset_provider_to_end(self):
    """Adding same name dataset through dataset provider to the end of the list."""

    class OriginalDataset(registered.RegisteredDataset):  # pylint: disable=unused-variable
      pass

    registered.add_dataset_builder_provider(
        TestBuilderProvider("original_dataset", EmptyDatasetBuilder),
        None,  # end of the list
    )
    builder_cls = registered.imported_builder_cls("original_dataset")
    self.assertEqual(builder_cls, OriginalDataset)

  def test_add_dataset_provider_to_start(self):
    """Adding same name dataset through dataset provider to the start of the list."""

    class MyDataset(registered.RegisteredDataset):  # pylint: disable=unused-variable
      pass

    registered.add_dataset_builder_provider(
        TestBuilderProvider("my_dataset", EmptyDatasetBuilder),
        0,  # start of the list
    )
    builder_cls = registered.imported_builder_cls("my_dataset")
    self.assertEqual(builder_cls, EmptyDatasetBuilder)

  def test_reset_dataset_builder_providers(self):
    """Resetting dataset builder providers."""

    class MyRegisteredDataset(registered.RegisteredDataset):  # pylint: disable=unused-variable
      pass

    registered.add_dataset_builder_provider(
        TestBuilderProvider("my_registered_dataset", EmptyDatasetBuilder),
        0,  # start of the list
    )
    builder_cls = registered.imported_builder_cls("my_registered_dataset")
    self.assertEqual(builder_cls, EmptyDatasetBuilder)

    registered.reset_dataset_builder_providers()

    registered.add_dataset_builder_provider(
        TestBuilderProvider("my_registered_dataset", EmptyDatasetBuilder),
        None,  # end of the list
    )
    builder_cls = registered.imported_builder_cls("my_registered_dataset")
    self.assertEqual(builder_cls, MyRegisteredDataset)


def test_skip_regitration():
  """Test `skip_registration()`."""

  with registered.skip_registration():

    class SkipRegisteredDataset(registered.RegisteredDataset):
      pass

  name = "skip_registered_dataset"
  assert name == SkipRegisteredDataset.name
  assert name not in load.list_builders()


def test_skip_regitration_kwarg():
  """Test `class Dataset(..., skip_registration=True)`."""

  class SkipDataset(registered.RegisteredDataset, skip_registration=True):
    pass

  name = "skip_dataset"
  assert name == SkipDataset.name
  assert name not in load.list_builders()


def test_custom_name():
  """Tests that builder can have custom names."""

  class SomeCustomNameBuilder(registered.RegisteredDataset):
    name = "custom_name"

  assert "custom_name" == SomeCustomNameBuilder.name
  assert "custom_name" in load.list_builders()


def test_name_inferred_from_pkg():
  ds_builder = dummy_ds_1_dataset_builder.Builder()
  assert ds_builder.name == "dummy_ds_1"


def test_name_inferred_from_pkg_level0_fails():
  pkg_path = (
      tfds.core.tfds_path() / "testing/dummy_config_based_datasets/dummy_ds_2"
  )
  expected_msg = (
      "When using `Builder` as class name, the dataset builder name is "
      'inferred from module name if named "*_dataset_builder" or from '
      'package name, but there is no package in "dummy_builder".'
  )
  with tfds.core.utils.add_sys_path(pkg_path):
    with pytest.raises(AssertionError, match=re.escape(expected_msg)):
      tfds.core.community.builder_cls_from_module("dummy_builder")


@mock.patch.dict(registered._DATASET_REGISTRY, {})
def test_name_inferred_from_pkg_level1():
  pkg_path = tfds.core.tfds_path() / "testing/dummy_config_based_datasets"
  with tfds.core.utils.add_sys_path(pkg_path):
    ds_builder = tfds.core.community.builder_cls_from_module(
        "dummy_ds_2.dummy_builder"
    )
  assert ds_builder.name == "dummy_ds_2"


@mock.patch.dict(registered._DATASET_REGISTRY, {})
def test_name_inferred_from_pkg_level2():
  pkg_path = tfds.core.tfds_path() / "testing"
  with tfds.core.utils.add_sys_path(pkg_path):
    ds_builder = tfds.core.community.builder_cls_from_module(
        "dummy_config_based_datasets.dummy_ds_2.dummy_builder"
    )
  assert ds_builder.name == "dummy_ds_2"


@mock.patch.dict(registered._DATASET_REGISTRY, {})
def test_name_inferred_from_pkg_level3():
  pkg_path = tfds.core.tfds_path()
  with tfds.core.utils.add_sys_path(pkg_path):
    ds_builder = tfds.core.community.builder_cls_from_module(
        "testing.dummy_config_based_datasets.dummy_ds_2.dummy_builder"
    )
  assert ds_builder.name == "dummy_ds_2"


class SourceDirDatasetBuilderProviderTest(testing.TestCase):

  def test_provider(self):
    provider = registered.SourceDirDatasetBuilderProvider(
        "testing/dummy_config_based_datasets"
    )
    self.assertEqual(
        set(provider.dataset_packages), {"dummy_ds_1", "dummy_ds_2"}
    )
    pkg_path, builder_module = provider.dataset_packages["dummy_ds_1"]
    self.assertEndsWith(
        str(pkg_path),
        "tensorflow_datasets/testing/dummy_config_based_datasets/dummy_ds_1",
    )
    self.assertEqual(
        builder_module,
        "tensorflow_datasets.testing.dummy_config_based_datasets.dummy_ds_1.dummy_ds_1_dataset_builder",
    )

  def test_imported_builder_cls(self):
    registered.add_dataset_builder_provider(
        registered.SourceDirDatasetBuilderProvider(
            "testing/dummy_config_based_datasets"
        )
    )
    builder = registered.imported_builder_cls("dummy_ds_1")
    self.assertEqual(builder.name, "dummy_ds_1")


# Tests for RegisteredDatasetCollection.


class EmptyDatasetCollectionBuilder(registered.RegisteredDatasetCollection):
  pass


class AbstractDatasetCollectionBuilder(registered.RegisteredDatasetCollection):

  @property
  @abc.abstractmethod
  def an_abstract_property(self):
    pass


def test_registered_dataset_collection():
  name = "empty_dataset_collection_builder"
  assert name == EmptyDatasetCollectionBuilder.name


def test_skip_dataset_collection_registration():
  with registered.skip_registration():

    class SkipCollectionBuilder(registered.RegisteredDatasetCollection):  # pylint: disable=unused-variable
      pass

  assert (
      "skip_collection_builder"
      not in registered.list_imported_dataset_collections()
  )


def test_duplicate_dataset_collection_builders():
  name = "duplicate_collection_builder"
  error_msg = f"DatasetCollection with name {name} already registered"

  class DuplicateCollectionBuilder(registered.RegisteredDatasetCollection):  # pylint: disable=unused-variable
    pass

  with pytest.raises(ValueError, match=error_msg):

    class DuplicateCollectionBuilder(registered.RegisteredDatasetCollection):  # pylint: disable=function-redefined
      pass


def test_duplicate_dataset_collections_on_notebooks():
  with mock.patch.object(py_utils, "is_notebook", lambda: True):
    name = "colab_dataset_collection_builder"
    assert name not in registered.list_imported_dataset_collections()

    class ColabDatasetCollectionBuilder(EmptyDatasetCollectionBuilder):  # pylint: disable=unused-variable
      pass

    assert name in registered.list_imported_dataset_collections()
    cls = registered.imported_dataset_collection_cls(name)
    assert isinstance(cls, type(ColabDatasetCollectionBuilder))
    old_colab_cls = ColabDatasetCollectionBuilder

    class ColabDatasetCollectionBuilder(EmptyDatasetBuilder):  # pylint: disable=function-redefined
      pass

    assert name in registered.list_imported_dataset_collections()
    assert isinstance(cls, type(ColabDatasetCollectionBuilder))
    assert not isinstance(cls, old_colab_cls)


def test_list_imported_dataset_collections():
  assert (
      "empty_dataset_collection_builder"
      in registered.list_imported_dataset_collections()
  )
  assert (
      "abstract_dataset_collection_builder"
      not in registered.list_imported_dataset_collections()
  )
  assert (
      "nonexistent_dc_builder"
      not in registered.list_imported_dataset_collections()
  )


def test_is_dataset_collection():
  assert registered.is_dataset_collection("empty_dataset_collection_builder")
  assert not registered.is_dataset_collection(
      "abstract_dataset_collection_builder"
  )
  assert not registered.is_dataset_collection("nonexistent_dc_builder")


def test_imported_dataset_collection_cls():
  existent = "empty_dataset_collection_builder"
  abstract = "abstract_dataset_collection_builder"
  nonexistent = "nonexistent_dataset_collection_builder"

  cls = registered.imported_dataset_collection_cls(existent)
  assert isinstance(cls, type(EmptyDatasetCollectionBuilder))

  error_msg = f"DatasetCollection {abstract} is an abstract class."
  with pytest.raises(AssertionError, match=error_msg):
    registered.imported_dataset_collection_cls(abstract)

  with pytest.raises(registered.DatasetCollectionNotFoundError):
    registered.imported_dataset_collection_cls(nonexistent)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataset feature for nested datasets."""

from __future__ import annotations

import dataclasses
import functools
from typing import Any, Callable, Dict, Iterator, Union

from tensorflow_datasets.core.data_sources import python
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.features import sequence_feature
from tensorflow_datasets.core.features import top_level_feature
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.core.utils.lazy_imports_utils import tree


@dataclasses.dataclass(frozen=True)
class _getitem:  # pylint: disable=invalid-name
  """A pickable version of getitem that can be fed to Beam pipelines."""

  decode_fn: Callable[[Any], Any]
  nest: Callable[[Any], Any]
  flat_example: list[Any]

  def __call__(self, i):
    return self.decode_fn(self.nest([v[i] for v in self.flat_example]))


class Dataset(sequence_feature.Sequence):
  """A Dataset feature encodes a nested dataset.

  `Dataset` corresponds to a dataset of `tfds.features.FeatureConnector`. Using
  `tfds.features.Dataset` will return a nested `tf.data.Dataset` inside the
  top-level `tf.data.Dataset` returned by `tfds.load`. At generation time, an
  iterable over the dataset elements is given.

  If you use tfds.data_source and the NumPy path, `Dataset` will return
  a [data
  source](https://www.tensorflow.org/datasets/api_docs/python/tfds/data_source).
  The advantage of having a data source is that decoding will be lazily executed
  when you access each example in the dataset.

  This is an experimental feature. Currently, only one level of nesting is
  supported and TF1 graph is not supported either.

  Example:
  At construction time (inside `_info`):

  ```python
    features=tfds.features.FeatureDict({
      'agent_id': np.object_,
      'episode': tfds.features.Dataset({
        'observation': tfds.features.Image(),
        'reward': tfds.features.Image(),
      }),
    })
  ```


  Will return:

  ```
  {
    'agent_id': tf.Tensor(shape=(), dtype=tf.string),
    'episode': tf.data.Dataset(element_spec={
        'observation': tf.Tensor(shape=(None, None, 3), dtype=tf.uint8),
        'reward': tf.Tensor(shape=(), dtype=tf.int32),
    }),
    }
  ```

  The nested dataset can be used as:

  ```
  for e in tfds.load(...): # {'agent_id': tf.Tensor, 'episode': tf.data.Dataset}
    for step in e['episode']:  # Each episode is a nested  `tf.data.Dataset`
      step['observation']
  ```

  During generation,  it accept any `Iterable`/`Iterator`, like

  ```python
  yield _, {
    'agent_id': agent_name
    'episode': ({'observation': ..., 'reward': ...} for _ in range(10)),
  }
  ```

  Or a dictionary of `Iterable`, like

  ```python
  yield _, {
    'agent_id': agent_name
    'episode': {'observation': np.ones(10), 'reward': np.ones(10)} ,
  }
  ```
  """

  # TODO(tfds): Add support for TF1 graph mode.

  @py_utils.memoize()
  def get_tensor_info(self):
    """Shape of one element of the dataset."""
    # Add the dataset level
    tensor_info = self._feature.get_tensor_info()
    return tree.map_structure(_add_dataset_lvl, tensor_info)

  def get_tensor_spec(self) -> tf.data.DatasetSpec:
    return tf.data.DatasetSpec(element_spec=self._feature.get_tensor_spec())

  @py_utils.memoize()
  def get_serialized_info(self):
    # Add the dataset level and the number of elements in the dataset
    tensor_info = super().get_serialized_info()
    return tree.map_structure(_add_dataset_lvl, tensor_info)

  def encode_example(
      self,
      example_ds: Union[Iterator[type_utils.TreeDict[Any]], Dict[str, Any]],
  ):
    if isinstance(example_ds, dict):
      dict_list = sequence_feature.transpose_dict_list(example_ds)
    else:
      dict_list = example_ds
    # Encode each individual element
    ds_elements = [
        self.feature.encode_example(example) for example in dict_list
    ]

    # Empty datasets return empty arrays
    if not ds_elements:
      return tree.map_structure(
          sequence_feature.build_empty_np, self.get_serialized_info()
      )

    # Then convert back list[nested dict] => nested dict[list]
    encoded = sequence_feature.stack_nested(ds_elements)
    return encoded

  def decode_example(self, serialized_example, decoders=None):
    # NOTE: By using from_tensor_slices we remove the possibility of nested
    # datasets.

    # Gets the decoding function of the inner feature to apply it to the
    # elements of the dataset.
    decode_fn = self.feature.decode_example
    if decoders:
      decode_fn = functools.partial(decode_fn, decoders=decoders)
    ds = tf.data.Dataset.from_tensor_slices(serialized_example).map(
        decode_fn, num_parallel_calls=tf.data.experimental.AUTOTUNE
    )
    return ds

  def decode_example_np(
      self, serialized_example, decoders=None
  ) -> python.PythonDataSource:
    """See base class for details."""
    flatten = self.feature._flatten  # pylint: disable=protected-access
    nest = self.feature._nest  # pylint: disable=protected-access
    flat_example = flatten(serialized_example)
    num_slices: int | None = None

    # Discover the number of slices in the Dataset (ie: the outter dimension).
    # We call slice a record in the Dataset.
    # We don't use `example` to avoid confusion with the `serialized_example`.
    for example_feature in flat_example:
      feature_num_slices = len(example_feature)
      if num_slices is None:
        num_slices = feature_num_slices
      else:
        if feature_num_slices != num_slices:
          raise ValueError(
              "The length of elements of all slices should be the same. Got"
              f" {num_slices} and {feature_num_slices}"
          )
    if num_slices is None:
      raise ValueError("no feature was found.")

    # Then, we apply the decode function on each slice.
    if isinstance(self.feature, top_level_feature.TopLevelFeature):
      # Only top-level features accept decoders.
      decode_fn = functools.partial(
          self.feature.decode_example_np, decoders=decoders
      )
    else:
      decode_fn = self.feature.decode_example_np

    return python.PythonDataSource(
        length=num_slices, getitem=_getitem(decode_fn, nest, flat_example)
    )

  def _flatten(self, x):
    """See base class for details."""
    return [x]

  def _nest(self, list_x):
    """See base class for details."""
    assert len(list_x) == 1

    return list_x[0]


def _add_dataset_lvl(tensor_info):
  """Add the dataset nesting level to the tensor_info."""
  tensor_info = feature_lib.TensorInfo.copy_from(tensor_info)
  tensor_info.dataset_lvl += 1
  return tensor_info



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""LJSpeech dataset."""

from __future__ import annotations

import os

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_URL = "https://keithito.com/LJ-Speech-Dataset/"
_DL_URL = "https://data.keithito.com/data/speech/LJSpeech-1.1.tar.bz2"


class Builder(tfds.core.GeneratorBasedBuilder):
  """LJSpeech dataset."""

  VERSION = tfds.core.Version("1.1.1")
  RELEASE_NOTES = {
      "1.1.1": "Fix speech data type with dtype=tf.int16.",
  }

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "id": np.str_,
            "speech": tfds.features.Audio(sample_rate=22050, dtype=np.int16),
            "text": tfds.features.Text(),
            "text_normalized": tfds.features.Text(),
        }),
        supervised_keys=("text_normalized", "speech"),
        homepage=_URL,
        metadata=tfds.core.MetadataDict(sample_rate=22050),
    )

  def _split_generators(self, dl_manager):
    extracted_dir = dl_manager.download_and_extract(_DL_URL)
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={"directory": extracted_dir},
        ),
    ]

  def _generate_examples(self, directory):
    """Yields examples."""
    metadata_path = os.path.join(directory, "LJSpeech-1.1", "metadata.csv")
    with epath.Path(metadata_path).open() as f:
      for line in f:
        line = line.strip()
        key, transcript, transcript_normalized = line.split("|")
        wav_path = os.path.join(
            directory, "LJSpeech-1.1", "wavs", "%s.wav" % key
        )
        example = {
            "id": key,
            "speech": wav_path,
            "text": transcript,
            "text_normalized": transcript_normalized,
        }
        yield key, example



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""imagenet_pi dataset."""

from tensorflow_datasets.datasets.imagenet_pi import imagenet_pi_dataset_builder
import tensorflow_datasets.public_api as tfds


class ImagenetPiTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for imagenet_pi dataset."""

  DATASET_CLASS = imagenet_pi_dataset_builder.Builder
  SPLITS = {
      'train': 100,  # Number of fake train example.
      'validation': 10,  # Number of fake validation example.
  }

  DL_EXTRACT_RESULT = {
      'labels_train': 'labels/train.csv',
      'labels_validation': 'labels/validation.csv',
      'confidences_train': 'confidences/train.csv',
      'confidences_validation': 'confidences/validation.csv',
      'train_path': 'ILSVRC2012_img_train.tar',
      'val_path': 'ILSVRC2012_img_val.tar',
  }

  SKIP_CHECKSUMS = True


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""To add retry logic to operations suceptible to transient failures."""

import random
import time
from typing import Callable, ParamSpec, TypeVar

from absl import logging
from tensorflow_datasets.core import constants


P = ParamSpec("P")
T = TypeVar("T")


def retry(func: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
  """Returns a decorator that retries the function."""
  # We purposely don't use flags, as this code might be run before flags are
  # parsed.
  tries = constants.TFDS_RETRY_TRIES
  delay = constants.TFDS_RETRY_INITIAL_DELAY
  multiplier = constants.TFDS_RETRY_DELAY_MULTIPLIER
  noise = constants.TFDS_RETRY_NOISE
  msg_substrings = constants.TFDS_RETRY_MSG_SUBSTRINGS
  for trial in range(1, tries + 1):
    try:
      return func(*args, **kwargs)
    except BaseException as err:  # pylint: disable=broad-except
      if trial >= tries:
        raise err
      msg = str(err)
      for msg_substring in msg_substrings:
        if msg_substring in msg:
          break
      else:
        raise err
      delay = delay + random.uniform(0, noise)
      logging.warning("%s, retrying in %s seconds...", msg, delay)
      time.sleep(delay)
      delay *= multiplier



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""multi_news dataset."""

from etils import epath
import tensorflow_datasets.public_api as tfds

_URL_PATH = (
    "https://huggingface.co/datasets/alexfabbri/multi_news/resolve/main/data/"
)
_LICENSE = "For non-commercial research and educational purposes only"


_DOCUMENT = "document"
_SUMMARY = "summary"


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for multi_news dataset."""

  VERSION = tfds.core.Version("2.1.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
      "2.0.0": "[Do not use] Update the dataset with valid URLs.",
      "2.1.0": (
          "Update the dataset with the correct URLs. The URLs in this version"
          " come from HuggingFace's dataset repo, which is curated by the same"
          " author: https://huggingface.co/datasets/alexfabbri/multi_news."
      ),
  }
  BLOCKED_VERSIONS = tfds.core.utils.BlockedVersions(
      versions={"2.0.0": "The URLs of this version are invalid."}
  )

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(
            {_DOCUMENT: tfds.features.Text(), _SUMMARY: tfds.features.Text()}
        ),
        supervised_keys=(_DOCUMENT, _SUMMARY),
        homepage="https://github.com/Alex-Fabbri/Multi-News",
        license=_LICENSE,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    data_dict = {
        "train_src": _URL_PATH + "train.src.cleaned",
        "train_tgt": _URL_PATH + "train.tgt",
        "val_src": _URL_PATH + "val.src.cleaned",
        "val_tgt": _URL_PATH + "val.tgt",
        "test_src": _URL_PATH + "test.src.cleaned",
        "test_tgt": _URL_PATH + "test.tgt",
    }
    files = dl_manager.download_and_extract(data_dict)
    return {
        "train": self._generate_examples(
            files["train_src"], files["train_tgt"]
        ),
        "validation": self._generate_examples(
            files["val_src"], files["val_tgt"]
        ),
        "test": self._generate_examples(files["test_src"], files["test_tgt"]),
    }

  def _generate_examples(self, src_file, tgt_file):
    """Yields examples."""
    with epath.Path(src_file).open() as src_f, epath.Path(
        tgt_file
    ).open() as tgt_f:
      for i, (src_line, tgt_line) in enumerate(zip(src_f, tgt_f)):
        yield i, {
            # In the original file, each line has one example and natural
            # newline tokens "\n" are being replaced with "NEWLINE_CHAR"
            # Here, we restore the natural newline token to avoid the special
            # vocab token "NEWLINE_CHAR".
            _DOCUMENT: src_line.strip().replace("NEWLINE_CHAR", "\n"),
            _SUMMARY: tgt_line.strip().lstrip(),
        }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""answer_equivalence dataset."""

from __future__ import annotations

import json
import os

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for answer_equivalence dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'qid': tfds.features.Text(),
            'gold_index': np.int32,
            'context': tfds.features.Text(),
            'question': tfds.features.Text(),
            'reference': tfds.features.Text(),
            'candidate': tfds.features.Text(),
            'score': np.float32,
            'question_1': tfds.features.ClassLabel(names=['no', 'yes', 'null']),
            'question_2': tfds.features.ClassLabel(names=['no', 'yes', 'null']),
            'question_3': tfds.features.ClassLabel(names=['no', 'yes', 'null']),
            'question_4': tfds.features.ClassLabel(names=['no', 'yes', 'null']),
        }),
        supervised_keys=None,
        homepage='https://github.com/google-research-datasets/answer-equivalence-dataset',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    homepage = 'https://raw.githubusercontent.com/google-research-datasets/answer-equivalence-dataset/main/v1/'
    archive = {
        'train': os.path.join(homepage, 'train.jsonl'),
        'ae_dev': os.path.join(homepage, 'ae_dev.jsonl'),
        'ae_test': os.path.join(homepage, 'ae_test.jsonl'),
        'dev_xlnet': os.path.join(homepage, 'dev_by_system/dev_xlnet.jsonl'),
        'dev_luke': os.path.join(homepage, 'dev_by_system/dev_luke.jsonl'),
        'dev_bidaf': os.path.join(homepage, 'dev_by_system/dev_bidaf.jsonl'),
    }

    paths = dl_manager.download_and_extract(archive)

    return {
        split: self._generate_examples(path) for split, path in paths.items()
    }

  def _generate_examples(self, filepath):
    """Yields examples."""
    with epath.Path(filepath).open() as fin:
      for i, line in enumerate(fin):
        data = json.loads(line)
        sample_id = data['qid'] + f'_{i}'
        sample = {
            'qid': data['qid'],
            'gold_index': data['gold_index'],
            'context': data['context'],
            'question': data['question'],
            'reference': data['reference'],
            'candidate': data['candidate'],
            'score': data['score'],
        }
        for j in range(4):
          sample[f'question_{j+1}'] = data['raw_rating'][j]
        yield sample_id, sample
